# meta sweep over jax_sweep.py

import argparse
import os
import subprocess
import time

import numpy as np

import matplotlib.pyplot as plt
import torch
from attrdict import AttrDefault, AttrDict
from collections import defaultdict

from domainbed import datasets

import json
import csv
import hashlib
from pathlib import Path

import pprint
pp = pprint.PrettyPrinter(indent=4)

from domainbed.scripts.sweep import make_args_list, Job

parser = argparse.ArgumentParser(description=' meta sweep over jax_sweep.py')
parser.add_argument('--dataset', type=str, default='PACS')
parser.add_argument('--action', type=str, default='train')
parser.add_argument('--n_trial', type=int, default=20)
parser.add_argument('--steps', type=int, default=20000)
parser.add_argument('--plot_freq', type=int, default=2000)
parser.add_argument('--checkpoint_freq', type=int, default=200)
parser.add_argument('--n_adapt', type=int, default=16)
parser.add_argument('--eta', type=float, default=100, help='ETA = eta * LR')
parser.add_argument('--methods', nargs='+', default=[])
parser.add_argument('--debug', action="store_true")
parser.add_argument('--MI', type=float, default=10., help='mutual information')
parser.add_argument('--REG', type=float, default=1., help='regularizer')
parser.add_argument('--GAP', type=float, default=0.1, help='Acc lower bound = min source domain acc - GAP')
parser.add_argument('--divdis_init', type=str, default='reinit', choices=['reinit', 'finetune'], help='divdis heads initialization method')

local_args = parser.parse_args()

# # get job output dirs
if local_args.dataset == "PACS":
    hparams="{\"N_WORKERS\": 4}"
    LEN_RECORDS = 18 
elif local_args.dataset == "ColoredMNIST01" or local_args.dataset == "ColoredMNIST":
    hparams="{\"expand_factor\": 8, \"N_WORKERS\": 4}"
    # interval: default (5000 steps / 100 plot_freq) + 1 = 51 records 
    LEN_RECORDS = 51 

def load_records(path):
    records = []
    for i, subdir in list(enumerate(os.listdir(path))):
        results_path = os.path.join(path, subdir, "results.jsonl")
        try:
            with open(results_path, "r") as f:
                for line in f:
                    records.append(json.loads(line[:-1]))
        except IOError:
            pass
    return records

## This part of code is to decide which pretrain records to use, i.e., to decide the best hparams of pretraining.

path = f'sweep_output/{local_args.dataset}/'
# len(records) = #hparams x #trial x #env 
records = load_records(path)[LEN_RECORDS-1::LEN_RECORDS] 
# len(results) = #hparams x #trial
results = defaultdict(list)
for r in records:
    if len(r['args']['test_envs']) == 1:
        test_env = r['args']['test_envs'][0]
        h_seed = r['args']['hparams_seed']
        t_seed = r['args']['trial_seed']
        train_val_acc = np.mean([r[k] for k in r if k.endswith('_out_acc') and k != f'env{test_env}_out_acc'])
        results[(h_seed, t_seed)].append((r['args']['output_dir'], r[f'env{test_env}_out_acc'], test_env, train_val_acc))

MAX_ACC = 0
MAX_SEED = None
for seed, r in results.items():
    # print(f"{seed=}:{len(r)=}") # assert_eq(len(r), #env)
    if sum([t[3] for t in r]) > MAX_ACC:
        MAX_ACC = sum([t[3] for t in r])
        MAX_SEED = seed
print(MAX_SEED)
print(results[MAX_SEED])
results[MAX_SEED] = sorted(results[MAX_SEED], key=lambda t: t[2])
job_output_dirs = [t[0] for t in results[MAX_SEED]]



# run jobs

from jax_save import jax_save
for job_output_dir in job_output_dirs:
    p = Path(job_output_dir) / 'embeddings'
    if (not p.exists()) or (not any(p.iterdir())):
        jax_save(job_output_dir)

for job_output_dir in job_output_dirs:
    command = f'''python jax_sweep.py \
                --output_dir {job_output_dir} \
                --n_trial {local_args.n_trial} \
                --steps {local_args.steps} \
                --plot_freq {local_args.plot_freq} \
                --checkpoint_freq {local_args.checkpoint_freq} \
                --n_adapt {local_args.n_adapt} \
                --eta {local_args.eta} \
                --methods {' '.join(local_args.methods)} \
                --MI {local_args.MI} \
                --REG {local_args.REG} \
                --GAP {local_args.GAP} \
                --divdis_init {local_args.divdis_init}
                '''
    print(command)
    if local_args.debug:
        pass 
    else:
        p = subprocess.Popen(command, shell=True)
        p.wait()







# report jobs results

jax_output_dirs = [f'{output_dir}/results_adapt:{local_args.n_adapt}_eta:{local_args.eta}'.replace(
    '.', '_') for output_dir in job_output_dirs]    
report_output_dir = f"sweep_output"
# (Pdb) jax_output_dirs
# ['sweep_output/PACS/58621fa2fd64aa671aa733c7c240ee76/results_adapt:256_eta:10_0', 'sweep_output/PACS/efad76522019e2d162d0ed76b9ddb07d/results_adapt:256_eta:10_0', 'sweep_output/PACS/9b0c22df4f6f2d919eaca9970da7d7d5/results_adapt:256_eta:10_0', 'sweep_output/PACS/511f6def0d9ed5117c646ee9178eb8de/results_adapt:256_eta:10_0']

methods = local_args.methods

# # records_list[i_job][method][trial][env][step]
records_list = [AttrDefault(lambda: [None] * local_args.n_trial) for _ in range(len(job_output_dirs))] 
for i_j in range(len(job_output_dirs)):
    for method in methods:
        for seed in range(local_args.n_trial):
            records_list[i_j][method][seed] = torch.load(
                f'{jax_output_dirs[i_j]}/{method}_trial{seed}_record.pt')
save_dicts = [torch.load(os.path.join(output_dir, 'model.pkl')) for output_dir in job_output_dirs]
args_list = [AttrDict(save_dict['args']) for save_dict in save_dicts]

n_bins = 10
test_env_names = [f'env{args.test_envs[0]}_out' for args in args_list]

# accs[method][job] = mean, 20 percentile, median, 80 percentile
def get_metric(metric_name):
    ret = [[] for _ in methods] 
    for i_m, method in enumerate(methods):
        for i_j in range(len(job_output_dirs)):
            ret[i_m].append([records_list[i_j][method][i][f'{test_env_names[i_j]}_{metric_name}'][local_args.steps // local_args.checkpoint_freq] for i in range(local_args.n_trial)])
        ret[i_m].append([
                    np.mean([records_list[i_j][method][i][f'{test_env_names[i_j]}_{metric_name}'][local_args.steps // local_args.checkpoint_freq] for i_j in range(len(job_output_dirs))])
                    for i in range(local_args.n_trial)
                ])
    erm = []
    for i_j in range(len(job_output_dirs)):
        erm.append(records_list[i_j][method][0][f'{test_env_names[i_j]}_{metric_name}'][0])
    erm.append(np.mean([records_list[i_j][method][0][f'{test_env_names[i_j]}_{metric_name}'][0] for i_j in range(len(job_output_dirs))]))
    return ret, erm
accs, acc_erm = get_metric('acc')
# eces, ece_erm = get_metric('ece')



from rich.console import Console
from rich.table import Table

report_path = (f'{report_output_dir}/{local_args.dataset}_N:{local_args.n_adapt}_M:{" ".join(local_args.methods)}_MI:{local_args.MI}_REG:{local_args.REG}_GAP:{local_args.GAP}'.replace('.', '_'))

def print_table_to_csv(table, file_path):
    '''add table content to csv file.'''
    n_cols = len(table.columns)
    n_rows = len(table.rows)
    with open(file_path, 'a') as f:
        writer = csv.writer(f)
        writer.writerow([])
        for i in range(n_rows):
            writer.writerow([table.columns[j]._cells[i] for j in range(n_cols)])

def print_table_to_txt(table, file_path):
    with open(file_path, 'at') as f:
        console = Console(file=f)
        console.print(table)

def print_table(table, file_path):
    print_table_to_csv(table, f'{file_path}_csv')
    print_table_to_txt(table, f'{file_path}_txt')

def table_add_row(table, method, *nums):
    num_strs = [f'{num:.3g}' for num in nums]
    table.add_row(method, *num_strs)

def create_table(metric_name, metrics, metric_erm):
    table = Table(title=metric_name)
    table.add_column("Method")
    for name in test_env_names:
        table.add_column(name)
    table.add_column("Mean")

    # ERM_accs = [t[1] for t in results[MAX_SEED]]
    # table_add_row(table, 'ERM', *ERM_accs, np.mean(ERM_accs))
    table_add_row(table, 'erm', *metric_erm)
    for i_m, method in enumerate(methods):
        table_add_row(table, method, *metrics[i_m])
    return table

def create_summary_table(metric_name, metrics, metric_20s, metric_erm):
    table = Table(title=f'{metric_name}_Summary')
    table.add_column("Method")
    table.add_column("Mean")
    table.add_column("Worst 20%")

    table_add_row(table, 'erm', np.mean(metric_erm), min(metric_erm))
    for i_m, method in enumerate(methods):
        table_add_row(table, method, np.mean(metrics[i_m]), min(metric_20s[i_m]))
    return table

acc_means = [[np.mean(acc) * 100 for acc in acc_] for acc_ in accs]
acc_20s = [[sorted(acc)[int(0.2*local_args.n_trial)] * 100 for acc in acc_] for acc_ in accs]
acc_50s = [[sorted(acc)[int(0.5*local_args.n_trial)] * 100 for acc in acc_] for acc_ in accs]
acc_80s = [[sorted(acc)[int(0.8*local_args.n_trial)] * 100 for acc in acc_] for acc_ in accs]
acc_erm = [a * 100 for a in acc_erm]

table_mean = create_table('acc_mean', acc_means, acc_erm)
table_20 = create_table('acc_20', acc_20s, acc_erm)
table_50 = create_table('acc_50', acc_50s, acc_erm)
table_80 = create_table('acc_80', acc_80s, acc_erm)
table_summary = create_summary_table('acc', acc_means, acc_20s, acc_erm)

for t in [table_mean, table_20, table_50, table_80, table_summary]:
    print_table(t, report_path)

