import argparse
import os
import subprocess
import time

import numpy as np

import matplotlib.pyplot as plt
import torch
from attrdict import AttrDefault, AttrDict
import GPUtil

from domainbed import datasets

parser = argparse.ArgumentParser(description='Jax train in output_dir')
parser.add_argument('--output_dir', type=str, default='train_output/PACS')
parser.add_argument('--action', type=str, default='train')
parser.add_argument('--n_trial', type=int, default=20)
parser.add_argument('--steps', type=int, default=20000)
parser.add_argument('--plot_freq', type=int, default=2000)
parser.add_argument('--checkpoint_freq', type=int, default=200)
parser.add_argument('--n_adapt', type=int, default=16)
parser.add_argument('--eta', type=float, default=100, help='ETA = eta * LR')
parser.add_argument('--methods', nargs='+', default=[])
parser.add_argument('--MI', type=float, default=10., help='mutual information')
parser.add_argument('--REG', type=float, default=1., help='regularizer')
parser.add_argument('--GAP', type=float, default=0.1, help='Acc lower bound = min source domain - GAP')
parser.add_argument('--divdis_init', type=str, default='reinit', choices=['reinit', 'finetune'], help='divdis heads initialization method')


local_args = parser.parse_args()
N_TRIAL = local_args.n_trial
output_dir = local_args.output_dir
plot_freq = local_args.plot_freq
checkpoint_freq = local_args.checkpoint_freq

jax_output_dir = f'{output_dir}/results_adapt:{local_args.n_adapt}_eta:{local_args.eta}'.replace(
    '.', '_')

save_dict = torch.load(os.path.join(output_dir, 'model.pkl'))
args = AttrDict(save_dict['args'])
hparams = AttrDict(save_dict['model_hparams'])

if args.dataset in vars(datasets):
    dataset = vars(datasets)[args.dataset](args.data_dir, args.test_envs,
                                           hparams)
else:
    raise NotImplementedError
n_envs = len(dataset.datasets)

# +
trained_params = AttrDefault(lambda: [None] * N_TRIAL)
methods = local_args.methods

available_gpus = GPUtil.getAvailable(limit=5)
n_gpus = len(available_gpus)
procs_by_gpu = [None] * n_gpus

commands = [(method, seed) for method in methods for seed in range(N_TRIAL)]
if local_args.action == 'train':
    while len(commands) > 0:
        for i_gpu, gpu in enumerate(available_gpus):
            proc = procs_by_gpu[i_gpu]
            if (proc is None) or (proc.poll() is not None):
                method, seed = commands.pop(0)
                command = f'''CUDA_VISIBLE_DEVICES={gpu} python jax_train.py \
                --output_dir {output_dir} \
                --method {method} \
                --eta {local_args.eta} \
                --n_adapt {local_args.n_adapt} \
                --load_from 0 \
                --steps {local_args.steps} \
                --checkpoint_freq {checkpoint_freq} \
                --MI {local_args.MI} \
                --REG {local_args.REG} \
                --GAP {local_args.GAP} \
                --divdis_init {local_args.divdis_init} \
                --seed {seed} 
                '''
                print(command)
                new_proc = subprocess.Popen(command, shell=True)
                procs_by_gpu[i_gpu] = new_proc
                break
        time.sleep(1)
    for p in procs_by_gpu:
        if p is not None:
            p.wait()